// This tab/file contains global variables and the setup routine

/*  This cube uses reversed polarity LED pairs to reduce the pincount for one layer from 64 down to 32.  This
    approach makes it possible to control and power the cube directly from the MEGA without the use of external
    shift registers.

    The timed interrupt or ISR for refreshing the cube is very efficient, meaning almost all the time involed is
    lighting LEDs.  This efficiency is derived from loading 4 bytes directly from the status arrays into the output
    ports.  We have to load those output ports 16 times during refresh - 8 layers with cathodes down and another 8
    layers with anodes down.  Refresh takes 8 msec. and we are refreshing every 10 msecs, so the MEGA spends 80% of
    its time refreshing.
*/

// content of the cube is stored in these two arrays
// they are stored in two different arrays to facilitate the transfer of whole bytes directly into output ports during refresh
volatile byte anodes[4][8];
volatile byte cathodes[4][8];
int xit=0;
// Next several items are used for rotation.
byte buffer_cube[8][8][8];
float myangle, myangle2, rotation; // used by the rotation routines
#define Black 0  // rotation code was originally for RGB cube
#define White 1

// *********************************Setup ******************************
void setup() {
  // set 48 pins to output and initialize low
  for (int x = 2; x < 14; x++) {
    pinMode(x, OUTPUT);
    digitalWrite(x, LOW);
  }
  for (int x = 18; x < 54; x++) {
    pinMode(x, OUTPUT);
    digitalWrite(x, LOW);
  }

  // TIMER 1 for interrupt frequency 80 Hz = refresh every 12.5 milliseconds
  cli(); // stop interrupts
  TCCR1A = 0; // set entire TCCR1A register to 0
  TCCR1B = 0; // same for TCCR1B
  TCNT1  = 0; // initialize counter value to 0
  // set compare match register for 80 Hz increments
  OCR1A = 24999; // = 16000000 / (8 * 80) - 1 (must be <65536)
  // turn on CTC mode
  TCCR1B |= (1 << WGM12);
  // Set CS12, CS11 and CS10 bits for 8 prescaler
  TCCR1B |= (0 << CS12) | (1 << CS11) | (0 << CS10);
  // enable timer compare interrupt
  TIMSK1 |= (1 << OCIE1A);
  sei(); // allow interrupts
  delay(1);
}
// ***************************** End Setup ************************************
