// This tab/file contains the sprite object class used for several animaitons and also a couple of useful tables.

/*  SPRITE OBJECT CLASS
  A sprite object has the following properties:
  .description: this is an array of minimum size 1,1,1 and max size 6,6,6.
  It sets the color (or lack of color) for each LED within the sprite. So it defines both the shape and color.
  It may be a single LED (1,1,1) or as much as 1/4 of the cube (6,6,6)
  .colorIt(color) is an alternative to building a description array. It turns the entire area of the sprite one color.
  .outline(color) is like colorIt but only colors the outline of your sprite
  .place: this is the location of the lower corner of the sprite.  It's a 3 number array specifying an X, Y, Z position.
  .motion: this is the direction the sprite will move when asked to move.  It's a 3 number array specifying the speed of movement in X,Y,Z directions.
  It will usually be a 0, +-1, or +-2 to get smooth movement within the cube.

  A sprite object also has some methods (or actions) it can take:
  .setIt:  this displays your spite in the cube at its specified place
  .clearIt:  this removes your sprite from the cube
  .moveIt: this moves your sprite one step in the direction specified by motion
  .bounceIt: similar to moveIt, but tests for cube boundaries and reverses course when an edge of the cube is detected.
  .rollX, rollY, rollZ: rolls the sprite around the edge of the cube in the specified direction.
  .rotateX, rotateY, rotateZ: rotates the the sprite on its own access in the specified direction.

  Sprites are created using the sprite constuctor, like this:
  sprite MySprite(1,2,3);
  where MySprite is the name of this particular instance of the object/class, and the 1,2,3 specifies its size in the X,Y,and Z directions.
  So a sprite could be a single LED, like:
  sprite LED1(1,1,1);
  The following is the definition of the sprite object class
  DO NOT MODIFY (or at least - Modify at your own risk!)
*/
class sprite {
  private:
    int lock;
    byte buffer[6][6][6];
    void copyBack() {  // copies the buffer content back into the description array
      for (int x = 0; x < 6; x++) {
        for (int y = 0; y < 6; y++) {
          for (int z = 0; z < 6; z++) {
            description[x][y][z] = buffer[x][y][z];
          }
        }
      }
    }
  public:
    int myX; // these are just the X,Y,and Z dimensions of a particular sprite instance.
    int myY;
    int myZ;
    // The following is a constuctor that lets us specify the dimensons of a sprite instance.
    // can be specified from 1,1,1 (default) up to 4,4,4 (but doesn't have to be a cube)
    sprite(int msizeX = 1, int msizeY = 1, int msizeZ = 1) {
      myX = msizeX;
      myY = msizeY;
      myZ = msizeZ;
    }

    // The following are the 3 properties of the sprite
    // Note: description is constructed 7-z,x,y, out of order
    // and with z reversed, to aid in builting it visually.  See examples to see why.
    byte description[6][6][6];  // describes its shape and color of a sprite instance.
    int place[3];  // specifies the lower corner of the position of the sprite
    int motion [3]; // specifies the direction it will move when asked to move
    // The following are the 4 methods or actions of the sprite class
    void setIt() { // this displays the sprite in the cube at it's specified place
      for (int x = 0; x < myX; x++) {
        for (int y = 0; y < myY; y++) {
          for (int z = 0; z < myZ; z++) {
            if (x + place[0] > -1 && x + place[0] < 8 && y + place[1] > -1 && y + place[1] < 8 && myZ - 1 - z + place[2] > -1 && myZ - 1 - z + place[2] < 8) {
              if (description[x][y][z] == 1) ledOn(x + place[0], y + place[1], myZ - 1 - z + place[2]);
              else ledOff(x + place[0], y + place[1], myZ - 1 - z + place[2]);
            }
          }
        }
      }
    }

    void clearIt() { // this clears it from the cube at its specified place
      for (int x = 0; x < myX; x++) {
        for (int y = 0; y < myY; y++) {
          for (int z = 0; z < myZ; z++) {
            ledOff(x + place[0], y + place[1], myZ - 1 - z + place[2]);
          }
        }
      }
    }
    // The next action is the easy way to describe a sprite.  lights the whole sprite.
    void fillIt() { // makes the whole description array one color
      for (int x = 0; x < 6; x++) {
        for (int y = 0; y < 6; y++) {
          for (int z = 0; z < 6; z++) {
            description[x][y][z] = 1;
          }
        }
      }
    }
    // This action is similar to fillIt, but only outline the sprite with the selected color.
    void outline(int mycolor) {
      for (int x = 0; x < myX; x++) {
        for (int y = 0; y < myY; y++) {
          for (int z = 0; z < myZ; z++) {
            description[x][y][z] = 1;
          }
        }
      }
      for (int x = 0; x < myX; x++) {
        for (int y = 1; y < myY - 1; y++) {
          for (int z = 1; z < myZ - 1; z++) {
            description[x][y][z] = 0;
          }
        }
      }
      for (int x = 1; x < myX - 1; x++) {
        for (int y = 0; y < myY; y++) {
          for (int z = 1; z < myZ - 1; z++) {
            description[x][y][z] = 0;
          }
        }
      }
      for (int x = 1; x < myX - 1; x++) {
        for (int y = 1; y < myY - 1; y++) {
          for (int z = 0; z < myZ; z++) {
            description[x][y][z] = 0;
          }
        }
      }
    }
    // This action is similar to colorIt, but creates a shere with the selected color.
    // Sprite should be a cube (all sides the same).
    // Also this feature is going to work best with large sprites like 6x6x6 or 5x5x5.
    void sphere() {
      float polar;
      for (int x = 0; x < myX; x++) {
        for (int y = 0; y < myY; y++) {
          for (int z = 0; z < myZ; z++) {
            float centerX = myX / 2;
            float centerY = myY / 2;
            float centerZ = myZ / 2;
            polar = sqrt((x - centerX) * (x - centerX) + (y - centerY) * (y - centerY) + (z - centerZ) * (z - centerZ)); // Calculate the distance
            if (polar < centerX) { // if an LED is inside the radius specified by myX, turn it on.
              description[x][y][z] = 1;
            }
            else {
              description[x][y][z] = 0;
            }
          }
        }
      }
    }
    void moveIt() {  // this moves it by one step in its specified direction of motion
      clearIt();
      place[0] = place[0] + motion[0];
      place[1] = place[1] + motion[1];
      place[2] = place[2] + motion[2];
      setIt();
    }

    void bounceIt() { // this moves it but also checks for edges of the cube.
      moveIt();     // if an edge is hit, direction in that one plane is reversed.
      if (place[0] < 1 && motion[0] < 0) {
        motion[0] = -motion[0];
      }
      if (place[1] < 1 && motion[1] < 0) {
        motion[1] = -motion[1];
      }
      if (place[2] < 1 && motion[2] < 0) {
        motion[2] = -motion[2];
      }
      if (place[0] > 7 - myX && motion[0] > 0) {
        motion[0] = -motion[0];
      }
      if (place[1] > 7 - myY && motion[1] > 0) {
        motion[1] = -motion[1];
      }
      if (place[2] > 7 - myZ && motion[2] > 0) {
        motion[2] = -motion[2];
      }
    }
    // for the next 6 actions, dir = 0 is clockwise rotation, while dir = 1 is counter-clockwise.
    void rollX(int dir) { // this rotates the sprite around the edge of the cube around X axis
      lock = 0;
      if (dir == 0) {
        if (motion[0] == 0 && motion[1] == 0 && motion[2] == -1) {
          lock = 1;
          if (place[2] < 1) {
            motion[0] = 0;
            motion[1] = -1;
            motion[2] = 0;
          }
        }
        if (motion[0] == 0 && motion[1] == -1 && motion[2] == 0) {
          lock = 1;
          if (place[1] < 1) {
            motion[0] = 0;
            motion[1] = 0;
            motion[2] = 1;
          }
        }
        if (motion[0] == 0 && motion[1] == 0 && motion[2] == 1) {
          lock = 1;
          if (place[2] > 7 - myZ) {
            motion[0] = 0;
            motion[1] = 1;
            motion[2] = 0;
          }
        }
        if (motion[0] == 0 && motion[1] == 1 && motion[2] == 0) {
          lock = 1;
          if (place[1] > 7 - myY) {
            motion[0] = 0;
            motion[1] = 0;
            motion[2] = -1;
          }
        }
        if (lock < 1 && place[0] < 7 - myZ) {
          motion[0] = 0; motion[1] = 0; motion[2] = -1;
        }
        if (lock < 1 && place[0] == 7 - myZ) {
          motion[0] = 0; motion[1] = -1; motion[2] = 0;
        }
      }
      else {
        if (motion[0] == 0 && motion[1] == 0 && motion[2] == 1) {
          lock = 1;
          if (place[2] > 7 - myZ) {
            motion[0] = 0;
            motion[1] = -1;
            motion[2] = 0;
          }
        }
        if (motion[0] == 0 && motion[1] == 1 && motion[2] == 0) {
          lock = 1;
          if (place[1] > 7 - myY) {
            motion[0] = 0;
            motion[1] = 0;
            motion[2] = 1;
          }
        }
        if (motion[0] == 0 && motion[1] == 0 && motion[2] == -1) {
          lock = 1;
          if (place[2] < 1) {
            motion[0] = 0;
            motion[1] = 1;
            motion[2] = 0;
          }
        }
        if (motion[0] == 0 && motion[1] == -1 && motion[2] == 0) {
          lock = 1;
          if (place[1] < 1) {
            motion[0] = 0;
            motion[1] = 0;
            motion[2] = -1;
          }
        }
        if (lock < 1 && place[2] < 7 - myZ) {
          motion[0] = 0; motion[1] = 0; motion[2] = 1;
        }
        if (lock < 1 && place[2] == 7 - myZ) {
          motion[0] = 0; motion[1] = 1; motion[2] = 0;
        }
      }
      moveIt();     // move on step
    }

    void rollY(int dir) { // this rotate the sprite around the edge of the cube around Y axis
      lock = 0;
      if (dir == 0) {
        if (motion[0] == 1 && motion[1] == 0 && motion[2] == 0) {
          lock = 1;
          if (place[0] > 7 - myX) {
            motion[0] = 0;
            motion[1] = 0;
            motion[2] = -1;
          }
        }
        if (motion[0] == 0 && motion[1] == 0 && motion[2] == 1) {
          lock = 1;
          if (place[2] > 7 - myZ) {
            motion[0] = 1;
            motion[1] = 0;
            motion[2] = 0;
          }
        }
        if (motion[0] == -1 && motion[1] == 0 && motion[2] == 0) {
          lock = 1;
          if (place[0] < 1) {
            motion[0] = 0;
            motion[1] = 0;
            motion[2] = 1;
          }
        }
        if (motion[0] == 0 && motion[1] == 0 && motion[2] == -1) {
          lock = 1;
          if (place[2] < 1) {
            motion[0] = -1;
            motion[1] = 0;
            motion[2] = 0;
          }
        }
        if (lock < 1 && place[0] < 7 - myX) { // used to be >1 which I think is a mistake  Changed to be consistent
          motion[0] = -1; motion[1] = 0; motion[2] = 0;
        }
        if (lock < 1 && place[0] == 7 - myX) { // used to be ==0
          motion[0] = 0; motion[1] = 0; motion[2] = -1;
        }
      }
      else {
        if (motion[0] == 1 && motion[1] == 0 && motion[2] == 0) {
          lock = 1;
          if (place[0] > 7 - myX) {
            motion[0] = 0;
            motion[1] = 0;
            motion[2] = 1;
          }
        }
        if (motion[0] == 0 && motion[1] == 0 && motion[2] == 1) {
          lock = 1;
          if (place[2] > 7 - myZ) {
            motion[0] = -1;
            motion[1] = 0;
            motion[2] = 0;
          }
        }
        if (motion[0] == -1 && motion[1] == 0 && motion[2] == 0) {
          lock = 1;
          if (place[0] < 1) {
            motion[0] = 0;
            motion[1] = 0;
            motion[2] = -1;
          }
        }
        if (motion[0] == 0 && motion[1] == 0 && motion[2] == -1) {
          lock = 1;
          if (place[2] < 1) {
            motion[0] = 1;
            motion[1] = 0;
            motion[2] = 0;
          }
        }
        if (lock < 1 && place[0] < 7 - myX) {
          motion[0] = 1; motion[1] = 0; motion[2] = 0;
        }
        if (lock < 1 && place[0] == 7 - myX) {
          motion[0] = 0; motion[1] = 0; motion[2] = 1;
        }
      }
      moveIt();     // move on step
    }

    void rollZ(int dir) { // this rotate the sprite around the edge of the cube around Z axis
      lock = 0;
      if (dir == 0) {
        if (motion[0] == 1 && motion[1] == 0 && motion[2] == 0) {
          lock = 1;
          if (place[0] > 7 - myX) {
            motion[0] = 0;
            motion[1] = -1;
            motion[2] = 0;
          }
        }
        if (motion[0] == 0 && motion[1] == 1 && motion[2] == 0) {
          lock = 1;
          if (place[1] > 7 - myY) {
            motion[0] = 1;
            motion[1] = 0;
            motion[2] = 0;
          }
        }
        if (motion[0] == -1 && motion[1] == 0 && motion[2] == 0) {
          lock = 1;
          if (place[0] < 1) {
            motion[0] = 0;
            motion[1] = 1;
            motion[2] = 0;
          }
        }
        if (motion[0] == 0 && motion[1] == -1 && motion[2] == 0) {
          lock = 1;
          if (place[1] < 1) {
            motion[0] = -1;
            motion[1] = 0;
            motion[2] = 0;
          }
        }
        if (lock < 1 && place[0] < 7 - myX) {
          motion[0] = 1; motion[1] = 0; motion[2] = 0;
        }
        if (lock < 1 && place[0] == 7 - myX) {
          motion[0] = 0; motion[1] = 1; motion[2] = 0;
        }
      }
      else {
        if (motion[0] == -1 && motion[1] == 0 && motion[2] == 0) {
          lock = 1;
          if (place[0] < 1) {
            motion[0] = 0;
            motion[1] = -1;
            motion[2] = 0;
          }
        }
        if (motion[0] == 0 && motion[1] == -1 && motion[2] == 0) {
          lock = 1;
          if (place[1] < 1) {
            motion[0] = 1;
            motion[1] = 0;
            motion[2] = 0;
          }
        }
        if (motion[0] == 1 && motion[1] == 0 && motion[2] == 0) {
          lock = 1;
          if (place[0] > 7 - myX) {
            motion[0] = 0;
            motion[1] = 1;
            motion[2] = 0;
          }
        }
        if (motion[0] == 0 && motion[1] == 1 && motion[2] == 0) {
          lock = 1;
          if (place[1] > 7 - myY) {
            motion[0] = -1;
            motion[1] = 0;
            motion[2] = 0;
          }
        }
        if (lock < 1 && place[0] < 7 - myX) {
          motion[0] = -1; motion[1] = 0; motion[2] = 0;
        }
        if (lock < 1 && place[0] == 7 - myX) {
          motion[0] = 0; motion[1] = -1; motion[2] = 0;
        }
      }
      moveIt();     // move on step
    }
    void rotateX(int dir) {
      for (int x = 0; x < 6; x++) {
        for (int y = 0; y < myY; y++) {
          for (int z = 0; z < myZ; z++) {
            if (dir == 0) {
              buffer[y][x][myZ - 1 - z] = description[z][x][y]; // clockwise rotatation around Z 90 degrees
            }
            else {
              buffer[myY - 1 - y][x][z] = description[z][x][y]; // counterclockwise rotatation 90 degrees
            }
          }
        }
      }
      copyBack(); // bring rotated object in buffer back to description
      setIt();  // show it
    }
    void rotateY(int dir) {
      for (int x = 0; x < myX; x++) {
        for (int y = 0; y < 6; y++) {
          for (int z = 0; z < myZ; z++) {
            if (dir == 0) {
              buffer[x][myZ - 1 - z][y] = description[z][x][y]; // clockwise rotatation around Y 90 degrees
            }
            else {
              buffer[myX - 1 - x][z][y] = description[z][x][y]; // counterclockwise rotatation 90 degrees
            }
          }
        }
      }
      copyBack();  // bring rotated object in buffer back to description
      setIt();  // show it
    }
    void rotateZ(int dir) {
      for (int x = 0; x < myX; x++) {
        for (int y = 0; y < myY; y++) {
          for (int z = 0; z < 6; z++) {
            if (dir == 0) {
              buffer[z][y][myX - 1 - x] = description[z][x][y]; // clockwise rotatation around Z 90 degrees
            }
            else {
              buffer[z][myY - 1 - y][x] = description[z][x][y]; // counterclockwise rotatation 90 degrees
            }
          }
        }
      }
      copyBack();  // bring rotated object in buffer back to description
      setIt();    // show it
    }
    void assignplace(unsigned char x, unsigned char y, unsigned char z)
    {
      place[0] = x;
      place[1] = y;
      place[2] = z;
    }

    void assignmotion(int x, int y, int z)
    {
      motion[0] = x;
      motion[1] = y;
      motion[2] = z;
    }
};   // end of class description


// This table gives us the distance of each LED from the center of the cube, for one layer.
float const distance[8][8] = {
  {4.949747468, 4.301162634, 3.807886553, 3.535533906, 3.535533906, 3.807886553, 4.301162634, 4.949747468,},
  {4.301162634, 3.535533906, 2.915475947, 2.549509757, 2.549509757, 2.915475947, 3.535533906, 4.301162634,},
  {3.807886553, 2.915475947, 2.121320344, 1.58113883, 1.58113883, 2.121320344, 2.915475947, 3.807886553,},
  {3.535533906, 2.549509757, 1.58113883, 0.707106781, 0.707106781, 1.58113883, 2.549509757, 3.535533906,},
  {3.535533906, 2.549509757, 1.58113883, 0.707106781, 0.707106781, 1.58113883, 2.549509757, 3.535533906,},
  {3.807886553, 2.915475947, 2.121320344, 1.58113883, 1.58113883, 2.121320344, 2.915475947, 3.807886553,},
  {4.301162634, 3.535533906, 2.915475947, 2.549509757, 2.549509757, 2.915475947, 3.535533906, 4.301162634,},
  {4.949747468, 4.301162634, 3.807886553, 3.535533906, 3.535533906, 3.807886553, 4.301162634, 4.949747468,}
};
// This table gives us coordinates to smoothly move from a side to the diagonal
byte const diagonal[8][8] = {
  {0, 0, 0, 0, 0, 0, 0, 0,        },
  {0, 0, 0, 0, 1, 1, 1, 1,        },
  {0, 0, 0, 1, 1, 2, 2, 2,        },
  {0, 0, 1, 1, 2, 2, 2, 3,        },
  {0, 1, 1, 2, 2, 3, 3, 4,        },
  {0, 1, 2, 2, 3, 3, 4, 5,        },
  {0, 1, 2, 3, 3, 4, 5, 6,        },
  {0, 1, 2, 3, 4, 5, 6, 7,        }
};
